package jaxt.examples.position;

import java.util.Random;

/**
 * Position indices on an 8*8 board. Note: this class has a natural ordering
 * that is inconsistent with equals.
 * We expect equals to be a congruence for the position abstraction.
 * 
 * @author Magne Haveraaen, 2007
 * 
 */
public class Position implements Comparable {

	/**
	 * Generates a dataset of 200 random elements, useful for testing. We could
	 * have chosen to create a complete test set of 64 positions, but this would
	 * not be optimal. For one thing, many axioms, such as transitivity of
	 * equality, need to be tested on distinct objects with the same value to
	 * have any real effect. Randomness on such a small domain ensures that we
	 * get some repeated values, and that these are spread out throughout the
	 * domain. A drawback of the chosen data set is that we do not generate
	 * values outside of the given range of the 8*8 board, and we do not check
	 * against superclass elements, e.g., Object.
	 * 
	 * @return data set of random values.
	 */
	public static Position[] positionDataSet() {
		Position[] data = new Position[200];
		Random g = new Random();
		for (int i = 0; i < data.length; i++) {
			data[i] = new Position(g.nextInt(8), g.nextInt(8));
		}
		return data;
	}

	/**
	 * data is given as an x-index and a y-index. These data should be within
	 * the range 0<=x,y<8.
	 */
	private int x, y;

	/**
	 * Constructs a position from two integers giving x- and y-values of index.
	 * Values outside of the board are folded into the board.
	 * 
	 * @param a
	 *            x-index
	 * @param b
	 *            y-index
	 */
	public Position(int a, int b) {
		x = a % 8;
		y = b % 8;
	}

	/**
	 * Ordering of posisitons based only the x-coordinate. Note: this class has
	 * a natural ordering that is inconsistent with equals.
	 * 
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Object obj) {
		final Position q = (Position) obj;
		return x - q.x;
	}

	/**
	 * Asserts that the data attributes are within required domain.
	 */
	public void dataInvariant() {
		assert 0 <= x;
		assert x < 8;
		assert 0 <= y;
		assert y < 8;
	}

	/**
	 * Two positions are equal if their x- and y-coordinates are equal,
	 * respectively.
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {
		final Position q = (Position) obj;
		return x == q.x && y == q.y;
	}

	/**
	 * @return x-coordinate
	 */
	public int getX() {
		return x;
	}

	/**
	 * @return y-coordinate
	 */
	public int getY() {
		return y;
	}

	/**
	 * The hashCode is reasonable well distributed. Would have been perfect if a
	 * larger factor than 8 was used.
	 * 
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		return 5 * x + y;
	}

	/**
	 * Moves this positions q places (by adding them together), folding the
	 * board if necesary. Can be used to move a chess piece, e.g., a knight
	 * along its movement pattern.
	 * 
	 * @param q
	 *            relative movement
	 */
	public void plus(Position q) {
		x = (x + q.x) % 8;
		y = (y + q.y) % 8;
	}

	/**
	 * String representation of position as a comma-separated list of
	 * coordinates inside paranthesis. The format is "(x,y)".
	 * 
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "(" + getX() + "," + getY() + ")";
	}
}
