package jaxt.java.lang;

import static junit.framework.Assert.*;

// import static java.lang.Integer.signum;

/**
 * Required axioms for the java interface java.lang.Comparable as taken from <a
 * url="http://java.sun.com/j2se/1.5.0/docs/api/">Java 5.0 API</a>
 * <p>
 * This interface imposes a total ordering on the objects of each class that
 * implements it. This ordering is referred to as the class's natural ordering,
 * and the class's compareTo method is referred to as its natural comparison
 * method.
 * <p>
 * These are axioms for the requirements on implementations of the interface.
 * See separate files for recommendations.
 * 
 * @see java.lang.Comparable
 * @author Magne Haveraaen & Karl Trygve Kalleberg, 2007
 * 
 */

abstract public class ComparableAxioms implements jaxt.framework.Axioms, Comparable {

	/**
	 * compareTo property 2: comparison with null.
	 * <p>
	 * e.compareTo(null) should throw a NullPointerException
	 */
	public static void compareToProperty2null(Comparable e) {
		try {
			e.compareTo(null);
			fail(e + ".compareTo(null) should throw a NullPointerException");
		} catch (NullPointerException npe) {
			// As required
		}
	}

	/**
	 * compareTo property 3: natural ordering is LE (less than or equal to),
	 * hence antisymmetric.
	 * <p>
	 * For the mathematically inclined, the relation that defines the natural
	 * ordering on a given class C is:
	 * <p align=center>
	 * {(x, y) such that x.compareTo((Object)y) <= 0}.
	 * </p>
	 */
	public static void compareToProperty3antisymmetricLE(Comparable x,
			Comparable y) {
		try {
			if (x.compareTo(y) <= 0 && y.compareTo(x) <= 0)
				assertEquals(x.compareTo(y), 0);
		} catch (RuntimeException e) {
			// some run-time exception occurred
		}
	}

	/**
	 * compareTo property 3: natural ordering is LE (less than or equal to),
	 * hence transitive.
	 * <p>
	 * For the mathematically inclined, the relation that defines the natural
	 * ordering on a given class C is:
	 * <p align=center>
	 * {(x, y) such that x.compareTo((Object)y) <= 0}.
	 * </p>
	 */
	public static void compareToProperty3transitiveLE(Comparable a,
			Comparable b, Comparable c) {
		try {
			if (a.compareTo(b) <= 0 && b.compareTo(c) <= 0)
				assertTrue(a.compareTo(c) <= 0);
		} catch (RuntimeException e) {
			// some run-time exception occurred
		}
	}

	/**
	 * compareTo property 3&4: the natural ordering is LE (less than or equal
	 * to) and an equivalence relation, hence reflexive.
	 * <p>
	 * For the mathematically inclined, the relation that defines the natural
	 * ordering on a given class C is:
	 * <p align=center>
	 * {(x, y) such that x.compareTo((Object)y) <= 0}.
	 * </p>
	 * <p>
	 * The quotient for this total order is:
	 * <p align=center>
	 * {(x, y) such that x.compareTo((Object)y) == 0}.
	 * </p>
	 */
	public static void compareToProperty4reflexive(Comparable x) {
		try {
			assertEquals(x.compareTo(x), 0);
		} catch (RuntimeException e) {
			// some run-time exception occurred
		}
	}

	/**
	 * compareTo property 4: the natural ordering is an equivalence relation,
	 * hence symmetric.
	 * <p>
	 * The quotient for this total order is:
	 * <p align=center>
	 * {(x, y) such that x.compareTo((Object)y) == 0}.
	 * </p>
	 */
	public static void compareToProperty4symmetricEQ(Comparable x, Comparable y) {
		try {
			if (x.compareTo(y) == 0)
				assertEquals(y.compareTo(x), 0);
		} catch (RuntimeException e) {
			// some run-time exception occurred
		}
	}

	/**
	 * compareTo property 4: the natural ordering is an equivalence relation,
	 * hence transitive.
	 * <p>
	 * The quotient for this total order is:
	 * <p align=center>
	 * {(x, y) such that x.compareTo((Object)y) == 0}.
	 * </p>
	 */
	public static void compareToProperty4transitiveEQ(Comparable x,
			Comparable y, Comparable z) {
		try {
			if (x.compareTo(y) == 0 && y.compareTo(z) == 0)
				assertEquals(x.compareTo(z), 0);
		} catch (RuntimeException e) {
			// some run-time exception occurred
		}
	}

	/**
	 * compareTo property 5: total order.
	 * <p>
	 * In the foregoing description, the notation sgn(expression) designates the
	 * mathematical signum function, which is defined to return one of -1, 0, or
	 * 1 according to whether the value of expression is negative, zero or
	 * positive. The implementor must ensure sgn(x.compareTo(y)) ==
	 * -sgn(y.compareTo(x)) for all x and y.
	 */
	public static void compareToProperty5totalOrder(Comparable x, Comparable y) {
		try {
			assertEquals(signum(x.compareTo(y)), -signum(y.compareTo(x)));
		} catch (RuntimeException e) {
			// some run-time exception occurred
		}
	}

	/**
	 * compareTo property 6: strong symmetry (both alternatives fail or both
	 * succeed).
	 * <p>
	 * This implies that x.compareTo(y) must throw an exception iff
	 * y.compareTo(x) throws an exception.
	 */
	public static void compareToProperty6strongSymmetry(Comparable x,
			Comparable y) {
		try {
			x.compareTo(y);
			y.compareTo(x);
			// neither call fails
		} catch (RuntimeException e) {
			// at least one of the calls throws an exception
			try {
				x.compareTo(y);
				fail(y + ".compareTo(" + x
						+ ") throws an exception while the converse does not");
			} catch (RuntimeException e1) {
				try {
					y.compareTo(x);
					fail(x
							+ ".compareTo("
							+ y
							+ ") throws an exception while the converse does not");
				} catch (RuntimeException e2) {
					// OK! Both calls fail symmetrically
				}
			}
		}
	}

	/**
	 * compareTo property 7: transitive GT (greater than).
	 * <p>
	 * The implementor must also ensure that the relation is transitive:
	 * (x.compareTo(y)>0 && y.compareTo(z)>0) implies x.compareTo(z)>0.
	 */
	public static void compareToProperty7transitiveGT(Comparable x,
			Comparable y, Comparable z) {
		try {
			if (x.compareTo(y) > 0 && y.compareTo(z) > 0)
				assertTrue(x.compareTo(z) > 0);
		} catch (RuntimeException e) {
			// some run-time exception occurred
		}
	}

	/**
	 * compareTo property 8: congruence wrt. ordering.
	 * <p>
	 * Finally, the implementer must ensure that x.compareTo(y)==0 implies that
	 * sgn(x.compareTo(z)) == sgn(y.compareTo(z)), for all z.
	 */
	public static void compareToProperty8congruenceCompareTo(Comparable x,
			Comparable y, Comparable z) {
		try {
			if (x.compareTo(y) == 0)
				assertTrue(signum(x.compareTo(z)) == signum(y.compareTo(z)));
		} catch (RuntimeException e) {
			// some run-time exception occurred
		}
	}

	public static int signum(int a) {
		return a < 0 ? -1 : a == 0 ? 0 : 1;
	}

}