package jaxt.java.lang;

import static junit.framework.Assert.*;
import java.util.*;
import static java.lang.StrictMath.abs;

/**
 * Recommendations for the java top level class java.lang.Object as taken from
 * <a url="http://java.sun.com/j2se/1.5.0/docs/api/">Java 5.0 API</a>
 * <p>
 * These are only recommendations. See separate files for axioms.
 * 
 * @see java.lang.Object
 * @author Magne Haveraaen & Karl Trygve Kalleberg, 2007
 * 
 */

abstract public class ObjectGlobalRecommendations extends Object implements jaxt.framework.GlobalAxioms {

	/**
	 * hashCode property 3: hash values should be well distributed.
	 * <p>
	 * It is not required that if two objects are unequal according to the
	 * equals(java.lang.Object) method, then calling the hashCode method on each
	 * of the two objects must produce distinct integer results. However, the
	 * programmer should be aware that producing distinct integer results for
	 * unequal objects may improve the performance of hashtables.
	 * 
	 * @param data
	 *            for which statistics is to be gathered
	 * @param reqDistinctHash
	 *            mimimum proportion of the hash values for distinct objects
	 *            that must be distinct
	 * @param reqMaximumSameHash
	 *            maximum proportion of reuse of one hash value
	 */
	public static void hashCodeProperty3wellDistribution(Object[] data,
			double reqDistinctHash, double reqMaximumSameHash) {
		assert reqDistinctHash <= 1.0;
		assert reqMaximumSameHash <= 1.0;
		Map<Integer, HashMap<Object, Integer>> stat = new HashMap<Integer, HashMap<Object, Integer>>();
		// gathering statistics
		for (Object d : data) {
			int hc = d.hashCode();
			HashMap<Object, Integer> h = stat.get(hc);
			if (h == null) {
				h = new HashMap<Object, Integer>();
				h.put(d, 0);
			} else {
				Integer i = h.get(d);
				if (i == null)
					i = 0;
				h.put(d, i + 1);
			}
			stat.put(hc, h);
		}
		// compute number of distinct hash values
		int noDistinctHash = stat.size();
		// compute maximum number of repeated hash values
		Collection<HashMap<Object, Integer>> c = stat.values();
		int noDistinctObjects = 0;
		int maxSameHash = 0;
		for (HashMap<Object, Integer> h : c) {
			noDistinctObjects += h.size();
			maxSameHash = maxSameHash < h.size() ? h.size() : maxSameHash;
		}
		// compute proportions
		double propDistinctHash = (double) noDistinctHash
				/ (double) noDistinctObjects;
		double propMaximumSameHash = (double) (maxSameHash - 1)
				/ (double) noDistinctHash;
		if (reqDistinctHash < 0)
			System.out
					.println("Proportion of data with distinct hash codes are "
							+ propDistinctHash + "=(" + noDistinctHash + "/"
							+ noDistinctObjects + ")");
		if (reqMaximumSameHash < 0)
			System.out
					.println("Proportion of reuse for most popular hash code is "
							+ propMaximumSameHash
							+ "=("
							+ (maxSameHash - 1)
							+ "/" + noDistinctHash + ")");
		assertFalse("Proportion of distinct hash codes is too low "
				+ propDistinctHash, propDistinctHash <= abs(reqDistinctHash));
		assertTrue("Proportion of reuse for hash code is to high "
				+ propMaximumSameHash,
				propMaximumSameHash <= abs(reqMaximumSameHash));
	}
}
